/*
    data.c - Part of psiconv, a PSION 5 data formats converter
    Copyright (c) 1999  Frodo Looijaard <frodol@dds.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include "data.h"
#include "list.h"
#include "general.h"

psiconv_cm_length_t psiconv_to_cm_length(psiconv_u32 len)
{
  return (2.54/1440.0) * ((psiconv_s32) len);
}

psiconv_point_size_t psiconv_to_point_size(psiconv_u32 len)
{
  return len / 20.0;
}

psiconv_bool_t psiconv_to_bool(psiconv_u8 val)
{
  return val?psiconv_bool_true:psiconv_bool_false;
}

psiconv_character_layout_status psiconv_basic_character_layout_status(void)
{
  struct psiconv_character_layout_status cl = 
  {
    { 0, 0, 0},            /* color               */
    { 0xff, 0xff, 0xff},   /* back_color          */
    10.0,                  /* font_size           */
    psiconv_bool_false,    /* italic              */
    psiconv_bool_false,    /* bold                */
    psiconv_normalscript,  /* super_sub           */
    psiconv_bool_false,    /* underline           */
    psiconv_bool_false,    /* crossed_out         */
    -1,                    /* font                */
  };

  return psiconv_clone_character_layout_status(&cl);
}


psiconv_paragraph_layout_status psiconv_basic_paragraph_layout_status(void)
{
  struct psiconv_paragraph_layout_status pl =
  {
    { 0xff, 0xff, 0xff},   /* back_color          */
    0.0,                   /* indent_left         */
    0.0,                   /* indent_right        */
    0.0,                   /* indent_first        */
    psiconv_justify_left,  /* justify             */
    0.0,                   /* interline           */
    psiconv_bool_false,    /* interline_exact     */
    0.0,                   /* top_space           */
    0.0,                   /* bottom_space        */
    psiconv_bool_false,    /* on_one_page         */
    psiconv_bool_false,    /* together_with       */
    psiconv_bool_false,    /* on_next_page        */
    psiconv_bool_false,    /* no_widow_protection */
    0.0,                   /* left_margin         */
    {                      /* bullet              */
      psiconv_bool_false,    /* on                  */
      psiconv_bool_true,     /* indent              */
      10.0,                  /* font_size           */
      -1,                    /* font                */
      0x95,                  /* character           */
      { 0,0,0}               /*  color              */
    },        
    {                      /* left_border         */
      psiconv_border_none,   /* kind                */
      1,                     /* thickness           */
      { 0,0,0 }              /* color               */
    },
    {                      /* right_border        */
      psiconv_border_none,   /* kind              */
      1,                     /* thickness         */
      { 0,0,0 }              /* color             */
    },
    {                      /* top_border          */
      psiconv_border_none,   /* kind                */
      1,                     /* thickness           */
      { 0,0,0 }              /* color               */
    },
    {                      /* bottom_border       */
      psiconv_border_none,   /* kind                */
      1,                     /* thickness           */
      { 0,0,0 }              /* color               */
    },
    {                      /* tabs            */
      {                      /* normal              */
        0.64,                  /* location            */
        psiconv_tab_left,      /* kind                */
      },
      NULL                   /* extras */
    }
  };
  psiconv_paragraph_layout_status res;
  
  pl.tabs.extras = psiconv_list_new(sizeof(psiconv_tab_t));
  res = psiconv_clone_paragraph_layout_status(&pl);
  psiconv_list_free(pl.tabs.extras);
  return res;
}
  
  
psiconv_character_layout_status psiconv_clone_character_layout_status
                                           (psiconv_character_layout_status ls)
{
  psiconv_character_layout_status result;

  result = malloc(sizeof(*result));
  *result = *ls;
  return result;
}

psiconv_paragraph_layout_status psiconv_clone_paragraph_layout_status
                                           (psiconv_paragraph_layout_status ls)
{
  psiconv_paragraph_layout_status result;

  result = malloc(sizeof(*result));
  *result = *ls;
  result->tabs.extras = psiconv_list_clone(result->tabs.extras);
  return result;
}

psiconv_style psiconv_get_style (psiconv_styles_section ss, int nr)
{
  if (nr == 0)
    return ss->normal;
  else
    return psiconv_list_get(ss->styles,0xff - nr);
}

psiconv_font_desc psiconv_get_font(psiconv_font_list fl, int nr)
{
  int i;
  psiconv_font_desc res;
  for (i = 0; i < psiconv_list_length(fl); i ++)  {
    res = psiconv_list_get(fl,i);
    if (res->id == nr)
      return res;
  }
  return NULL;
}

