/*
    data.h - Part of psiconv, a PSION 5 data formats converter
    Copyright (c) 1999  Frodo Looijaard <frodol@dds.nl>       

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* This file contains the declarations of all types that are used to
   represent the Word file. Variables of these types are written by the
   parse routines, and read by the generation routines. */

#ifndef PSICONV_DATA_H
#define PSICONV_DATA_H

#include "general.h"
#include "list.h"

typedef struct psiconv_flagged_string
{
  char *data;
  psiconv_u8 flags;
} *psiconv_flagged_string;

typedef float psiconv_cm_length_t;    /* For offsets in centimeters */

typedef float psiconv_point_size_t;   /* For sizes in points */

typedef enum psiconv_bool
{
  psiconv_bool_false,
  psiconv_bool_true, 
} psiconv_bool_t;

typedef enum psiconv_super_sub
{ psiconv_normalscript,
  psiconv_superscript,
  psiconv_subscript
} psiconv_super_sub_t;

typedef enum psiconv_justify
{ psiconv_justify_left, 
  psiconv_justify_centre, 
  psiconv_justify_right, 
  psiconv_justify_full 
} psiconv_justify_t;

typedef struct psiconv_color
{
  psiconv_u8 red;
  psiconv_u8 green;
  psiconv_u8 blue;
} psiconv_color_t;

typedef struct psiconv_bullet
{
  psiconv_bool_t on;
  psiconv_bool_t indent;
  psiconv_point_size_t font_size;
  int font;
  psiconv_u8 character;
  psiconv_color_t color;
} psiconv_bullet_t;

typedef enum psiconv_border_kind
{ psiconv_border_none,
  psiconv_border_solid,
  psiconv_border_double,
  psiconv_border_dotted,
  psiconv_border_striped,
  psiconv_border_dotstripe,
  psiconv_border_dotdotstripe
} psiconv_border_kind_t;

typedef struct psiconv_border
{ 
  psiconv_border_kind_t kind;
  psiconv_u8 thickness;
  psiconv_color_t color;
} psiconv_border_t;

typedef enum psiconv_tab_kind
{
  psiconv_tab_left,
  psiconv_tab_centre,
  psiconv_tab_right
} psiconv_tab_kind_t;

typedef struct psiconv_tab
{
  psiconv_cm_length_t location;
  psiconv_tab_kind_t kind;
} psiconv_tab_t;


typedef psiconv_list psiconv_tab_list; /* of tab_t */

typedef struct psiconv_all_tabs
{
  psiconv_tab_t normal;
  psiconv_tab_list extras;
} psiconv_all_tabs_t;

typedef struct psiconv_character_layout_status
{
  psiconv_color_t color;
  psiconv_color_t back_color;
  psiconv_point_size_t font_size;
  psiconv_bool_t italic;
  psiconv_bool_t bold;
  psiconv_super_sub_t super_sub;
  psiconv_bool_t underline;
  psiconv_bool_t crossed_out;
  int font; /* number as in font list */
} *psiconv_character_layout_status;

typedef struct psiconv_paragraph_layout_status
{
  psiconv_color_t back_color;
  psiconv_cm_length_t indent_left;
  psiconv_cm_length_t indent_right;
  psiconv_cm_length_t indent_first;
  psiconv_justify_t justify;
  psiconv_point_size_t interline;
  psiconv_bool_t interline_exact;
  psiconv_point_size_t top_space;
  psiconv_point_size_t bottom_space;
  psiconv_bool_t on_one_page;
  psiconv_bool_t together_with; /* What the heck is this? */
  psiconv_bool_t on_next_page;
  psiconv_bool_t no_widow_protection;
  psiconv_cm_length_t left_margin;
  psiconv_bullet_t bullet;
  psiconv_border_t left_border;
  psiconv_border_t right_border;
  psiconv_border_t top_border;
  psiconv_border_t bottom_border;
  psiconv_all_tabs_t tabs;
} *psiconv_paragraph_layout_status;

typedef struct psiconv_style
{
  psiconv_character_layout_status character;
  psiconv_paragraph_layout_status paragraph;
  psiconv_u8 hotkey;
  psiconv_flagged_string name;
  psiconv_u32 unknown;
} *psiconv_style;

typedef struct psiconv_header_section
{
  psiconv_u32 uid1;
  psiconv_u32 uid2;
  psiconv_u32 uid3;
  psiconv_u32 checksum;
  psiconv_u32 sectiontable_offset;
} *psiconv_header_section;

typedef struct psiconv_sectiontable_entry
{
  psiconv_u32 id;
  psiconv_u32 offset;
} *psiconv_sectiontable_entry;

typedef psiconv_list psiconv_sectiontable_section; /* Of sectiontable_entry */

typedef struct psiconv_status_section
{
  psiconv_u8 unknown;
  psiconv_bool_t show_tabs;
  psiconv_bool_t show_spaces;
  psiconv_bool_t show_paragraph_ends;
  psiconv_bool_t show_line_breaks;;
  psiconv_bool_t show_hard_minus;;
  psiconv_bool_t show_hard_space;;
  psiconv_bool_t show_full_pictures;
  psiconv_bool_t show_full_graphs;
  psiconv_bool_t show_top_toolbar;
  psiconv_bool_t show_side_toolbar;
  psiconv_bool_t fit_lines_to_screen;
  psiconv_u32 cursor_position;
  psiconv_u32 display_size;
} *psiconv_status_section;

typedef struct psiconv_appl_id_section
{
  psiconv_u32 id;
  psiconv_flagged_string name;
} *psiconv_appl_id_section;

typedef struct psiconv_in_line_layout
{
  psiconv_character_layout_status layout;
  int length;
} *psiconv_in_line_layout;

typedef psiconv_list psiconv_in_line_layouts; /* of in_line_layout */

typedef struct psiconv_paragraph
{
  char *text;
  psiconv_character_layout_status base_character;
  psiconv_paragraph_layout_status base_paragraph;
  psiconv_s16 base_style;
  psiconv_in_line_layouts in_lines;
} *psiconv_paragraph;

typedef psiconv_list psiconv_text_and_layout; /* Of paragraphs */

typedef psiconv_list psiconv_style_list; /* Of style */

typedef struct psiconv_styles_section
{
  psiconv_style normal;
  psiconv_style_list styles;
} *psiconv_styles_section;

typedef struct psiconv_header
{
  psiconv_bool_t on_first_page;
  psiconv_text_and_layout paragraphs;
} *psiconv_header;

typedef struct psiconv_page_section
{
  psiconv_u32 first_page_nr;
  psiconv_cm_length_t header_dist;
  psiconv_cm_length_t footer_dist;
  psiconv_cm_length_t left_margin;
  psiconv_cm_length_t right_margin;
  psiconv_cm_length_t top_margin;
  psiconv_cm_length_t bottom_margin;
  psiconv_cm_length_t page_width;
  psiconv_cm_length_t page_height;
  psiconv_header header;
  psiconv_header footer;
} * psiconv_page_section;

typedef struct psiconv_font_desc
{
  char *name;
  psiconv_u8 id;
} *psiconv_font_desc;

typedef psiconv_list psiconv_font_list; /* of font_desc */

typedef struct psiconv_word_file {
  psiconv_page_section page_sec;
  psiconv_font_list fonts;
  psiconv_text_and_layout paragraphs;
  psiconv_status_section status_sec;
  psiconv_styles_section styles_sec;
} *psiconv_word_file;

#define PSICONV_ID_PSION_5 0x10000037
#define PSICONV_ID_DATA_FILE 0x1000006D
#define PSICONV_ID_WORD_FILE 0x1000007F

#define PSICONV_ID_STATUS_SECTION 0x10000243
#define PSICONV_ID_APPL_ID_SECTION 0x10000089
#define PSICONV_ID_TEXT_SECTION 0x10000106
#define PSICONV_ID_LAYOUT_SECTION 0x10000143
#define PSICONV_ID_STYLES_SECTION 0x10000104
#define PSICONV_ID_PAGE_SECTION 0x10000105

/* Convert to cm_length_t */
extern psiconv_cm_length_t psiconv_to_cm_length(psiconv_u32 len);

/* Convert to point_size_t */
extern psiconv_point_size_t psiconv_to_point_size(psiconv_u32 len);

/* Convert to bool_t */
extern psiconv_bool_t psiconv_to_bool(psiconv_u8 val);


/* Return a clean layout_status. You can modify it at will */
extern psiconv_character_layout_status 
                                 psiconv_basic_character_layout_status(void);

/* Return a clean layout_status. You can modify it at will */
extern psiconv_paragraph_layout_status 
                                  psiconv_basic_paragraph_layout_status(void);

/* Clone a layout_status: the new copy is completely independent of the
   original one */

extern psiconv_paragraph_layout_status psiconv_clone_paragraph_layout_status
                                           (psiconv_paragraph_layout_status ls);

extern psiconv_character_layout_status psiconv_clone_character_layout_status
                                          (psiconv_character_layout_status ls);

/* Get a numbered style */
extern psiconv_style psiconv_get_style (psiconv_styles_section ss, int nr);

extern psiconv_font_desc psiconv_get_font(psiconv_font_list fl, int nr);


#endif /* def PSICONV_DATA_H */
