/*
    parser.c - Part of psiconv, a PSION 5 data formats converter
    Copyright (c) 1999  Frodo Looijaard <frodol@dds.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Parse the PSION Word file */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "data.h"
#include "general.h"
#include "parser.h"
#include "main.h"

static void fatal(int level,psiconv_u32 off, const char *format,...);
static void progress(int level,psiconv_u32 off, const char *format,...);
static void warn(int level,psiconv_u32 off, const char *format,...);
static void debug(int level,psiconv_u32 off, const char *format,...);

static char *make_printable(const char *s);

static psiconv_u8 read_u8(const psiconv_buffer buf,int lev, psiconv_u32 off);
static psiconv_u16 read_u16(const psiconv_buffer buf,int lev, psiconv_u32 off);
static psiconv_u32 read_u32(const psiconv_buffer buf,int lev, psiconv_u32 off);

static psiconv_cm_length_t read_cm_length(const psiconv_buffer buf, int lev,
                                          psiconv_u32 off, int *length);
static psiconv_point_size_t read_point_size
                  (const psiconv_buffer buf, int lev,psiconv_u32 off, 
                   int *length);
static psiconv_flagged_string read_string(const psiconv_buffer buf,int lev,
                                          psiconv_u32 off,int *length);
static psiconv_color_t read_color_t(const psiconv_buffer buf, int lev,
                                    psiconv_u32 off, int *length);

static int parse_font(const psiconv_buffer buf, int lev, psiconv_u32 off, 
                      int *len, int *font, psiconv_font_list fl);
static int parse_border(const psiconv_buffer buf,int lev,psiconv_u32 off,
                        int *len,psiconv_border_t *border);
static int parse_header_section(const psiconv_buffer buf,int lev,
                                psiconv_u32 off, int *length, 
                                psiconv_header_section *h);
static int parse_sectiontable(const psiconv_buffer buf, int lev,psiconv_u32 off,
                              int *length,psiconv_sectiontable_section *s);
static int parse_status_section(const psiconv_buffer buf, int lev,
                               psiconv_u32 off, int *length, 
                               psiconv_status_section *s);
static int parse_appl_id_section(const psiconv_buffer buf, int lev, 
                                 psiconv_u32 off, int *length,
                                 psiconv_appl_id_section *a);
static int parse_text_section(const psiconv_buffer buf,int lev,psiconv_u32 off,
                              int *length,psiconv_text_and_layout *t);
static int parse_paragraph_layout_codes(const psiconv_buffer buf, int lev,
                                        psiconv_u32 off, int *len, 
                                        psiconv_paragraph_layout_status stat,
                                        psiconv_font_list fl);
static int parse_character_layout_codes(const psiconv_buffer buf, int lev,
                                        psiconv_u32 off, int *len, 
                                        psiconv_character_layout_status stat, 
                                        psiconv_font_list fl);
static int parse_styles_section(const psiconv_buffer buf, int lev, 
                                psiconv_u32 off, int *length,
                                psiconv_styles_section *ss, 
                                psiconv_font_list fl);
static void no_layout_section(psiconv_text_and_layout t,
                              psiconv_styles_section ss);
static int parse_layout_section(const psiconv_buffer buf,int lev,
                                psiconv_u32 off, int *length,
                                psiconv_text_and_layout t,
                                psiconv_font_list fl,psiconv_styles_section ss);
static int parse_header_footer(const psiconv_buffer buf,int lev,
                               psiconv_u32 off, int *length,
                               psiconv_header *hd,psiconv_font_list fl);
static int parse_page_section(const psiconv_buffer buf,int lev,psiconv_u32 off,
                       int *length, psiconv_page_section *ps, 
                       psiconv_font_list fl);
static psiconv_s64 offset_in_sectiontable(psiconv_u32 id,
                                          const psiconv_sectiontable_section s);


void fatal(int level, psiconv_u32 off, const char *format,...)
{
  va_list ap;
  va_start(ap,format);
  fprintf(stderr,"Fatal error (offset %08x): ",off);
  vfprintf(stderr,format,ap);
  fputs("\n",stderr);
  va_end(ap);
  exit(1);
}

void progress(int level, psiconv_u32 off, const char *format,...)
{
  int i;
  va_list ap;
  va_start(ap,format);
  if (psiconv_verbosity >= PSICONV_VERB_PROGRESS) {
    fprintf(stderr,"%08x ",off);
    for (i = 0; i < level; i++)
      fputs("=",stderr);
    fprintf(stderr,"> ");
    vfprintf(stderr,format,ap);
    fputs("\n",stderr);
  }
  va_end(ap);
}

void warn(int level, psiconv_u32 off, const char *format,...) 
{
  va_list ap;
  va_start(ap,format);
  if (psiconv_verbosity >= PSICONV_VERB_WARN) {
    fprintf(stderr,"WARNING (offset %08x): ",off);
    vfprintf(stderr,format,ap);
    fputs("\n",stderr);
  }
  va_end(ap);
}

void debug(int level, psiconv_u32 off, const char *format,...)
{
  int i;
  va_list ap;
  va_start(ap,format);
  if (psiconv_verbosity >= PSICONV_VERB_DEBUG) {
    fprintf(stderr,"%08x ",off);
    for (i = 0; i < level; i++)
      fputs("-",stderr);
    fprintf(stderr,"> ");
    vfprintf(stderr,format,ap);
    fputs("\n",stderr);
  }
  va_end(ap);
}

char *make_printable(const char *s)
{
  int i;
  char *res = malloc(strlen(s) + 1);
  for (i = 0; i < strlen(s); i ++)
    if (s[i] < 0x20 || s[i] >= 0x7f) 
      res[i] = '.';
    else
      res[i] = s[i];
  res[strlen(s)] = 0;
  return res;
}


psiconv_u8 read_u8(const psiconv_buffer buf,int lev,psiconv_u32 off)
{
  psiconv_u8 *ptr;
  ptr = psiconv_list_get(buf,off);
  if (!ptr)
    fatal(lev,off,"Trying byte read past the end of the file");
  return *ptr;
}

psiconv_u16 read_u16(const psiconv_buffer buf,int lev,psiconv_u32 off)
{
  psiconv_u8 *ptr0,*ptr1;
  ptr0 = psiconv_list_get(buf,off);
  ptr1 = psiconv_list_get(buf,off+1);
  if (!ptr0 || !ptr1)
    fatal(lev,off,"Trying word read past the end of the file");
  return *ptr0 + (*ptr1 << 8);
}

psiconv_u32 read_u32(const psiconv_buffer buf,int lev,psiconv_u32 off)
{
  psiconv_u8 *ptr0,*ptr1,*ptr2,*ptr3;
  ptr0 = psiconv_list_get(buf,off);
  ptr1 = psiconv_list_get(buf,off+1);
  ptr2 = psiconv_list_get(buf,off+2);
  ptr3 = psiconv_list_get(buf,off+3);
  if (!ptr0 || !ptr1 || !ptr2 || !ptr3)
    fatal(lev,off,"Trying long read past the end of the file");
  return *ptr0 + (*ptr1 << 8) + (*ptr2 << 16) + (*ptr3 << 24);
}


psiconv_cm_length_t read_cm_length(const psiconv_buffer buf, int lev,
                                   psiconv_u32 off, int *length)
{
  psiconv_cm_length_t res;
  if (length)
    *length = 4;
  res = psiconv_to_cm_length(read_u32(buf,lev,off));
  debug(lev+1,off,"Length: %f",res);
  return res;
}

psiconv_point_size_t read_point_size(const psiconv_buffer buf, int lev,
                                     psiconv_u32 off, int *length)
{
  psiconv_point_size_t res;
  if (length)
    *length = 4;
  res = psiconv_to_point_size(read_u32(buf,lev,off));
  debug(lev+1,off,"Size: %f",res);
  return res;
}

psiconv_flagged_string read_string(const psiconv_buffer buf,int lev,
                                   psiconv_u32 off,int *length)
{
  psiconv_u8 len;
  int i;
  psiconv_flagged_string result;
  char *res_copy;
  
  progress(lev+1,off,"Going to read a flagged string");

  result = malloc(sizeof(*result));
  len = read_u8(buf,lev,off);
  result->flags = len & 0x03;
  len = len >> 2;
  debug(lev+2,off,"Flags: %01x",result->flags);
  debug(lev+2,off,"Length: %d",len);

  result->data = malloc(len + 1);
  for (i = 0; i < len; i++)
    result->data[i] = read_u8(buf,lev,off+1+i);
  result->data[len] = 0;

  res_copy = make_printable(result->data);
  debug(lev+2,off,"Contents: `%s'",res_copy);
  free(res_copy);
  len ++;

  if (length)
    *length = len;

  progress(lev+1,off+len-1,"End of string (total length: %08x)",len);

  return result;
}

psiconv_color_t read_color_t(const psiconv_buffer buf, int lev,
                             psiconv_u32 off, int *length)
{
  psiconv_color_t result;
  result.red = read_u8(buf,lev,off);
  result.green = read_u8(buf,lev,off+1);
  result.blue = read_u8(buf,lev,off+2);
  debug(lev+1,off,"Color: red %02x, green %02x, blue %02x",result.red, 
        result.green, result.blue);
  if (length)
    *length = 3;
  return result;
}

int parse_font(const psiconv_buffer buf, int lev, psiconv_u32 off, int *length, 
               int *font, psiconv_font_list fl)
{
  int res = 0;
  int strlength,i;
  char *str;
  char *str_copy;
  char *str_copy2;
  psiconv_font_desc el;
  psiconv_u8 id;
  int len;

  progress(lev+1,off,"Going to parse font");
  *font = -1;
  strlength = read_u8(buf,lev+2,off);
  str = malloc(strlength);
  for (i = 0; i < strlength-1; i++) 
    str[i] = read_u8(buf,lev+2,off + 1 + i);
  str[strlength-1] = 0;
  id = read_u8(buf,lev+2,off + strlength);
  str_copy = make_printable(str);
  debug(lev+2,off+1,"Found font `%s', id %02x",str_copy,id);
  free(str_copy);

  len = strlength + 1;
  if (length)
    *length = len;
  
  for (i = 0; (*font == -1) && (i < psiconv_list_length(fl)); i ++) {
    el = psiconv_list_get(fl,i);
    if (el->id == id)  {
      if (! strcmp(el->name,str)) {
        *font = id;
        goto FOUND;
      } else {
        warn(lev+2,off,"Two fonts with same id, but different names!");
        str_copy = make_printable(str);
        str_copy2 = make_printable(el->name);
        debug(lev,off,"id: %02x, in list: `%s', now: `%s'",id,str_copy2, 
              str_copy);
        free(str_copy);
        free(str_copy2);
        res = -1;
        goto FOUND;
      }
    } else if (! strcmp(el->name,str)) {
      str_copy = make_printable(str);
      warn(lev+2,off,"Two fonts with same name, but different ids!");
      debug(lev+2,off,"font: %s, in list: %02x, now: %02x",str_copy,el->id,id);
      free(str_copy);
      res = -1;
      goto FOUND;
    }
  }
  
  el = malloc(sizeof(*el));
  el->name = str;
  el->id = id;
  *font = id;
  psiconv_list_add(fl,el);
  free(el);
  goto END;
  return res;

FOUND:
  free(str);

END:
  progress(lev+1,off + len - 1,"End of font (total length: %08x)",len);
  return res;
}


int parse_border(const psiconv_buffer buf,int lev,psiconv_u32 off,int *length,
                 psiconv_border_t *border)
{
  int res = 0;
  psiconv_u32 temp;
  int leng,i;
  int len = 0;

  progress(lev+1,off,"Going to parse border data");
  progress(lev+2,off+len,"Going to read border kind");
  temp = read_u8(buf,lev+2,off+len);
  if (temp == 0x00)
    border->kind = psiconv_border_none;
  else if (temp == 0x01)
    border->kind = psiconv_border_solid;
  else if (temp == 0x02)
    border->kind = psiconv_border_double;
  else if (temp == 0x03)
    border->kind = psiconv_border_dotted;
  else if (temp == 0x04)
    border->kind = psiconv_border_striped;
  else if (temp == 0x05)
    border->kind = psiconv_border_dotstripe;
  else if (temp == 0x06)
    border->kind = psiconv_border_dotdotstripe;
  else {
    warn(lev+2,off,"Unknown border kind (defaults to `none')");
    border->kind = psiconv_border_none;
    res = -1;
  }
  debug(lev+2,off+len,"Kind: %02x",temp);
  len ++;

  progress(lev+2,off+len,"Going to read border thickness");
  temp = read_u8(buf,lev+2,off+len);
  if (!(border->kind == psiconv_border_solid || 
        border->kind == psiconv_border_double ||
        temp == 0x00 || temp == 0x01)) {
    warn(lev+2,off,"Border thickness specified for unlikely border type");
    res = -1;
  }
  border->thickness = temp;
  debug(lev+2,off+len,"Thickness: %02x",temp);
  len ++;

  progress(lev+2,off+len,"Going to read three unknown bytes "
                         "(0x00 0x00 0x00 expected)");
  for (i = 0; i < 2; i++) {
    temp = read_u8(buf,lev,off + len + i);
    if (temp != 0x00) {
      warn(lev+2,off,"Unknown byte in border specification");
      debug(lev+2,off+len+i,"Byte %d: read %02x, expected %02x",i,temp,0x00);
      res = -1;
    }
  }
  len += 3;

  progress(lev+2,off+len,"Going to read the border color");
  border->color = read_color_t(buf,lev+2,off+len,&leng);
  len += leng;
  
  progress(lev+2,off+len,"Going to read the final unknown byte "
                         "(0x01 expected)");
  temp = read_u8(buf,lev+2,off + len);
  if (temp != 0x01) {
    warn(lev+2,off,"Unknown last byte in border specification");
    debug(lev+2,off+len,"Last byte: read %02x, expected %02x",temp,0x01);
    res = -1;
  }
  len ++;

  if (length)
    *length = len;
  
  progress(lev+1,off + len - 1,
           "End of border (total length: %08x)",len);

  return res;
}

int parse_header_section(const psiconv_buffer buf,int lev,psiconv_u32 off,
                         int *length, psiconv_header_section *h)
{
  int res=0;
  int len=0;

  progress(lev+1,off,"Going to read the header section");
  *h = malloc(sizeof (**h));

  progress(lev+2,off,"Going to read UID1 (0x%08x expected)",PSICONV_ID_PSION_5);
  (*h)->uid1 = read_u32(buf,lev+2,off+len);
  if ((*h)->uid1 != PSICONV_ID_PSION_5) {
    warn(lev+2,off+len,"This may not be a Psion-5 document. It certainly is "
                       "no Word document.");
    debug(lev+2,off+len,"UID1: read %08x, expected %08x",(*h)->uid1,
          PSICONV_ID_PSION_5);
    res = -1;
  }
  len += 0x04;

  progress(lev+2,off+len,"Going to read UID2 (0x%08x expected)",
           PSICONV_ID_DATA_FILE);
  (*h)->uid2 = read_u32(buf,lev+2,off+len);
  if ((*h)->uid2 != PSICONV_ID_DATA_FILE) {
    warn(lev+2,off+len,"This may not be a document file. It certainly is "
                       "no Word document.");
    debug(lev+2,off+len,"UID2: read %08x, expected %08x",(*h)->uid2,
          PSICONV_ID_DATA_FILE);
    res = -1;
  }
  len += 0x04;

  progress(lev+2,off+len,"Going to read UID3 (0x%08x expected)",
           PSICONV_ID_WORD_FILE);
  (*h)->uid3 = read_u32(buf,lev+2,off+len);
  if ((*h)->uid3 != PSICONV_ID_WORD_FILE) {
    warn(lev+2,off+len,"This is not be a Word document");
    debug(lev+2,off+len,"UID3: read %08x, expected %08x",(*h)->uid3,
          PSICONV_ID_WORD_FILE);
    res = -1;
  }
  len += 0x04;

  progress(lev+2,off+len,"Going to read Checksum (0x55089ffe expected)");
  (*h)->checksum = read_u32(buf,lev+2,off+len);
  if ((*h)->checksum != 0x55089ffe) {
    warn(lev+2,off+len,"Checksum failed (ignore if one of the UIDs was off)");
    debug(lev+2,off+len,"Checksum: read %08x, expected %08x",
          (*h)->checksum,0x55089ffe);
    res = -1;
  }
  len += 0x04;

  progress(lev+2,off+len,"Going to read Section Table Offset");
  (*h)->sectiontable_offset = read_u32(buf,lev+2,off+len);
  debug(lev+2,off+len,"Section Table Offset: %08x", (*h)->sectiontable_offset);

  progress(lev+1,off+len-1,"End of header section");

  if (length)
    *length = len;

  return res;
}

int parse_sectiontable(const psiconv_buffer buf, int lev,psiconv_u32 off,
                       int *length,psiconv_sectiontable_section *s)
{
  int i;
  psiconv_u8 nr;
  int res=0;
  int len=0;
  psiconv_sectiontable_entry ste;

  progress(lev+1,off+len,"Going to read the section table section");
  *s = psiconv_list_new(sizeof(*ste));

  progress(lev+2,off+len,"Going to read the section table length");
  nr = read_u8(buf,lev+2,off+len);
  debug(lev+2,off+len,"Length: %08x",nr);
  if (nr & 0x01) {
    warn(lev+2,off+len,"Section table length odd - ignoring last entry");
    res = -1;
  }
  len ++;

  progress(lev+2,off+len,"Going to read the section table entries");
  ste = malloc(sizeof(*ste));
  for (i = 0; i < nr / 2; i++) {
    ste->id = read_u32(buf,lev+2,off + len);
    debug(lev+2,off + len,"Entry %d: ID = %08x",i,ste->id);
    len += 0x04;
    ste->offset = read_u32(buf,lev+2,off + len);
    debug(lev+2,off +len,"Entry %d: Offset = %08x",i,ste->offset);
    psiconv_list_add(*s,ste);
    len += 0x04;
  }

  free(ste);

  progress(lev+1,off+len-1,"End of section table");

  if (length)
    *length = len;

  return res;
}

int parse_status_section(const psiconv_buffer buf, int lev,psiconv_u32 off,
                        int *length, psiconv_status_section *s)
{
  int res=0;
  int len=0;
  psiconv_u32 temp;

  progress(lev+1,off,"Going to read the status section");
  *s = malloc(sizeof(**s));

  progress(lev+2,off+len,"Going to read first byte (0x%02x expected)",0x02);
  (*s)->unknown = read_u8(buf,lev+2,off+len);
  if ((*s)->unknown != 0x02) {
    res = -1;
    warn(lev+2,off+len,"First byte read of status section failed. "
                       "Its use is unknown.");
    debug(lev+2,off+len,
          "First byte: read %02x, expected %02x",(*s)->unknown,0x02);
  }
  len += 0x01;

  progress(lev+2,off+len,"Going to read first byte of display flags");
  temp = read_u8(buf,lev+2,off+len);

  (*s)->show_tabs = psiconv_to_bool(temp & 0x01);
  debug(lev+2,off+len,"Show tabs: %02x",(*s)->show_tabs);
  (*s)->show_spaces = psiconv_to_bool((temp >> 1)& 0x01);
  debug(lev+2,off+len,"Show spaces: %02x",(*s)->show_spaces);
  (*s)->show_paragraph_ends = psiconv_to_bool((temp >> 2)& 0x01);
  debug(lev+2,off+len,"Show paragraph ends: %02x",(*s)->show_paragraph_ends);
  (*s)->show_line_breaks = psiconv_to_bool((temp >> 3)& 0x01);
  debug(lev+2,off+len,"Show line breaks: %02x",(*s)->show_line_breaks);
  (*s)->show_hard_minus = psiconv_to_bool((temp >> 5)& 0x01);
  debug(lev+2,off+len,"Show hard minus: %02x",(*s)->show_hard_minus);
  (*s)->show_hard_space = psiconv_to_bool((temp >> 6)& 0x01);
  debug(lev+2,off+len,"Show hard space: %02x",(*s)->show_hard_space);
  if (temp & 0x90) {
    warn(lev+2,off+len,"First byte of display flags in status sections "
                       "contains unknown flags");
    debug(lev+2,off+len,"Unknown flags: %02x",temp & 0x90);
    res = -1;
  }
  len ++;

  progress(lev+2,off+len,"Going to read second byte of display flags");
  temp = read_u8(buf,lev+2,off+len);

  (*s)->show_full_pictures = psiconv_to_bool(temp & 0x01);
  debug(lev+2,off+len,"Show full pictures: %02x",(*s)->show_full_pictures);
  (*s)->show_full_graphs = psiconv_to_bool((temp >> 1)& 0x01);
  debug(lev+2,off+len,"Show full graphs: %02x",(*s)->show_full_graphs);
  if (temp & 0xfc) {
    warn(lev+2,off+len,"Second byte of display flags in status section "
                       "contains unknown flags");
    debug(lev+2,off+len,"Unknown flags: %02x",temp & 0xfc);
    res = -1;
  }
  len ++;

  progress(lev+2,off+len,"Going to read top toolbar setting");
  temp = read_u8(buf,lev+2,off+len);

  if (temp == 0x00 || temp == 0x01) {
    (*s)->show_top_toolbar = psiconv_to_bool(temp);
    debug(lev+2,off+len,"Show top toolbar: %02x",temp);
  }
  else {
    warn(lev+2,off+len,"Unknown data in show top toolbar setting in status "
                       "section (defaults to on)");
    debug(lev+2,off+len,"Unknown top toolbar value: %02x",temp);
    res = -1;
    (*s)->show_top_toolbar = psiconv_bool_true;
  }
  len ++;

  progress(lev+2,off+len,"Going to read side toolbar setting");
  temp = read_u8(buf,lev+2,off+len);

  if (temp == 0x00 || temp == 0x01) {
    (*s)->show_side_toolbar = psiconv_to_bool(temp);
    debug(lev+2,off,"Show side toolbar: %02x",temp);
  }
  else {
    warn(lev+2,off+len,"Unknown data in show side toolbar setting in status "
                       "section (defaults to on)");
    debug(lev+2,off,"Unknown side toolbar value: %02x",temp);
    res = -1;
    (*s)->show_side_toolbar = psiconv_bool_true;
  }
  len ++;
  
  progress(lev+2,off+len,"Going to read operational flags");
  temp = read_u8(buf,lev+2,off+len);

  (*s)->fit_lines_to_screen = psiconv_to_bool((temp >> 3) & 0x08);
  debug(lev+2,off+len,"Fit lines to screen: %02x",(*s)->fit_lines_to_screen);
  if (temp & 0xf7) {
    warn(lev+2,off+len,"Operational flags byte in status section "
                       "contains unknown flags");
    debug(lev+2,off+len,"Unknown flags: %02x",temp & 0xfc);
    res = -1;
  }
  len ++;
  
  progress(lev+2,off+len,"Going to read cursor position");
  (*s)->cursor_position = read_u32(buf,lev+2,off + len);
  debug(lev+2,off+len,"Cursor position: %08x",(*s)->cursor_position);
  len += 0x04;

  progress(lev+2,off+len,"Going to read display size");
  (*s)->cursor_position = read_u32(buf,lev+2,off + len);
  debug(lev+2,off+len,"Display size: %08x",(*s)->cursor_position);
  len += 0x04;

  if (length)
    *length = len;

  progress(lev+1,off+len-1,"End of status table");

  return res;
}


int parse_appl_id_section(const psiconv_buffer buf, int lev, psiconv_u32 off, 
                          int *length,psiconv_appl_id_section *a)
{
  int res=0;
  int len=0;
  int leng;

  progress(lev+1,off,"Going to read the application id section");
  *a = malloc(sizeof(**a));

  progress(lev+2,off+len,"Going to read the type identifier (%08x expected)",
           PSICONV_ID_WORD_FILE);
  (*a)->id = read_u32(buf,lev+2,off+len);
  if ((*a)->id != PSICONV_ID_WORD_FILE) {
    debug(lev+2,off+len,"Type: read %08x, expected %08x",(*a)->id,
          PSICONV_ID_WORD_FILE);
    warn(lev+2,off+len,
         "Type field in application id section indicates a non-word document.");
  }
  len += 0x04;

  progress(lev+2,off+len,"Going to read the application id string "
                          "(`Word.app' expected)");
  (*a)->name = read_string(buf,lev+2,off + len,&leng);

  len += leng;
  progress(lev+1,off+len-1,"End of appliction id section");

  if (length)
    *length = len;

  return res;
}

int parse_text_section(const psiconv_buffer buf,int lev,psiconv_u32 off,
                       int *length,psiconv_text_and_layout *t)
{
  psiconv_u32 text_len;
  psiconv_paragraph para;

  int nr;
  int res = 0;
  int i,j,start;
  int len=0;
  char *str_copy;
 
  progress(lev+1,off,"Going to parse the text section");
  progress(lev+2,off,"Reading the text length");
  text_len = read_u8(buf,lev+2,off);
  if ((text_len & 0x03) == 3) {
    text_len = read_u32(buf,lev+2,off) >> 3;
    len += 4;
  } else if ((text_len & 0x03) == 1) {
    text_len = read_u16(buf,lev+2,off) >> 2;
    len += 2;
  } else {
    text_len = text_len >> 1;
    len += 1;
  }
  debug(lev+2,off,"Length: %08x",text_len);

  *t = psiconv_list_new(sizeof(*para));
  para = malloc(sizeof(*para));

  progress(lev+2,off+len,"Going to read all paragraph text");
  nr = 0;
  start = 0;
  for (i = 0; i < text_len; i++)
    if (read_u8(buf,lev+2,off+len+i) == 0x06) {
      para->text = malloc(i - start + 1);
      for (j = 0; j < i - start; j++)
        para->text[j] = read_u8(buf,lev+1,off + len + start + j);
      para->text[j] = 0;
      psiconv_list_add(*t,para);

      str_copy = make_printable(para->text);
      debug(lev+2,off+i+len,"Line %d: %d characters",nr,strlen(str_copy)+1);
      debug(lev+2,off+i+len,"Line %d: `%s'",nr,str_copy);
      free(str_copy);

      start = i + 1;
      nr ++;
    }
  
  if (start != text_len) {
    res = -1;
    warn(lev+2,off+start+len,
         "Last line does not end on EOL (%d characters left)", len - start);
    para->text = malloc(text_len - start + 1);
    for (j = 0; j < text_len - start; j++)
      para->text[j] = read_u8(buf,lev+2,off + start + j + len);
    para->text[text_len - start] = 0;
    psiconv_list_add(*t,para);
    str_copy = make_printable(para->text);
    debug(lev+2,off+start+len,"Last line: %d characters",nr,strlen(str_copy)+1);
    debug(lev+2,off+start+len,"Last line: `%s'",str_copy);
    free(str_copy);
  }
 
  free(para);

  len += text_len;

  progress(lev+1,off+len-1,"End of text section");

  if (length)
    *length = len;

  return res;
}

int parse_paragraph_layout_codes(const psiconv_buffer buf, int lev,
                                 psiconv_u32 off, int *len, 
                                 psiconv_paragraph_layout_status stat,
                                 psiconv_font_list fl)
{
  int res=0;
  int nr,i,leng;
  psiconv_u32 temp;
  psiconv_u8 id,bul_leng;
  psiconv_tab_t tab;

  progress(lev+1,off,"Going to read paragraph layout codes");
  nr = read_u32(buf,lev+2,off);
  debug(lev+2,off,"Length: %08x",nr);
  for (i = 0; i < nr; i++) {
    id = read_u8(buf,lev+2,off + 4 + i);
    switch(id) {
      case 0x01:
        progress(lev+2,off + 4 + i,"Going to read background color");
        stat->back_color = read_color_t(buf,lev+2,off+5+i,&leng);
        i += leng;
        break;
      case 0x02:
        progress(lev+2,off + 4 + i,"Going to read indent left");
        stat->indent_left = read_cm_length(buf,lev+2,off + 5 + i,&leng);
        debug(lev+2,off + 5 + i,"Indent left: %6.3f",stat->indent_left);
        i += leng;
        break;
      case 0x03:
        progress(lev+2,off + 4 + i,"Going to read indent right");
        stat->indent_right = read_cm_length(buf,lev+2,off + 5 + i,&leng);
        debug(lev+2,off + 5 + i,"Indent right: %6.3f",stat->indent_right);
        i += leng;
        break;
      case 0x04:
        progress(lev+2,off + 4 + i,"Going to read indent first");
        stat->indent_first = read_cm_length(buf,lev+2,off + 5 + i, &leng);
        debug(lev+2,off + 5 + i,"Indent first: %6.3f",stat->indent_first);
        i += leng;
        break;
      case 0x05:
        progress(lev+2,off + 4 + i,"Going to read justify");
        temp = read_u8(buf,lev+2,off+5+i);
        if (temp == 0x00)
          stat->justify = psiconv_justify_left;
        else if (temp == 0x01)
          stat->justify = psiconv_justify_centre;
        else if (temp == 0x02)
          stat->justify = psiconv_justify_right;
        else if (temp == 0x03)
          stat->justify = psiconv_justify_full;
        else {
          res = -1;
          warn(lev+2,off + 5 + i,
               "Unknown justify argument in paragraph layout codes list");
        }
        debug(lev+2,off + 5 + i,"Justify: %02x",temp);
        i ++;
        break;
      case 0x07:
        progress(lev+2,off + 4 + i,"Going to read interline");
        stat->interline = read_point_size(buf,lev+2,off + 5 + i,&leng);
        debug(lev+2,off + 5 + i,"Interline: %6.3f",stat->interline);
        i += leng;
        break;
      case 0x08:
        progress(lev+2,off + 4 + i,"Going to read interline exact");
        temp = read_u8(buf,lev+2,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->interline_exact = psiconv_to_bool(temp);
        else {
          warn(lev+2,off + 5 + i,
               "Unknown interline_exact argument in pararaph layout codes "
               "list");
          res = -1;
        }
        debug(lev+2,off + 5 + i,"Interline exact: %02x",temp);
        i ++;
        break;
      case 0x09:
        progress(lev+2,off + 4 + i,"Going to read top space");
        stat->top_space = read_point_size(buf,lev+1,off + 5 + i,&leng);
        debug(lev+2,off + 5 + i,"Top space: %6.3f",stat->top_space);
        i += leng;
        break;
      case 0x0a:
        progress(lev+2,off + 4 + i,"Going to read bottom space");
        stat->bottom_space = read_point_size(buf,lev+2,off + 5 + i,&leng);
        debug(lev+2,off + 5 + i,"Bottom space: %6.3f",stat->bottom_space);
        i += leng;
        break;
      case 0x0b:
        progress(lev+2,off + 4 + i,"Going to read on one page");
        temp = read_u8(buf,lev+2,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->on_one_page = psiconv_to_bool(temp);
        else  {
          res = -1;
          warn(lev+2,off + 5 + i,
               "Unknown on_one_page argument in pararaph layout codes list");
        }
        debug(lev+2,off + 5 + i,"On one page: %02x",temp);
        i ++;
        break;
      case 0x0c:
        progress(lev+2,off + 4 + i,"Going to read together with");
        temp = read_u8(buf,lev+2,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->together_with = psiconv_to_bool(temp);
        else {
          res = -1;
          warn(lev+2,off + 5 + i,
               "Unknown together_with argument in pararaph layout codes list");
        }
        debug(lev+2,off + 5 + i,"Together with: %02x",temp);
        i ++;
        break;
      case 0x0d:
        progress(lev+2,off + 4 + i,"Going to read on next page");
        temp = read_u8(buf,lev+2,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->on_next_page = psiconv_to_bool(temp);
        else {
          warn(lev+2,off + 5 + i,
               "Unknown on_next_page argument in pararaph layout codes list");
          res = -1;
        }
        debug(lev+2,off + 5 + i,"On next page: %02x",temp);
        i ++;
        break;
      case 0x0e:
        progress(lev+2,off + 4 + i,"Going to read no widow protection");
        temp = read_u8(buf,lev+2,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->no_widow_protection = psiconv_to_bool(temp);
        else  {
          warn(lev+2,off + 5 + i,
               "Unknown no_widow_protection argument in pararaph layout codes "
               "list");
          res = -1;
        }
        debug(lev+2,off + 5 + i,"No widow protection: %02x",temp);
        i ++;
        break;
      case 0x10:
        progress(lev+2,off + 4 + i,"Going to read left margin");
        stat->left_margin = read_cm_length(buf,lev+2,off + 5 + i,&leng);
        debug(lev+2,off + 5 + i,"Left margin: %6.3f",stat->left_margin);
        i += leng;
        break;
      case 0x11:
        progress(lev+2,off + 4 + i,"Going to read border top");
        res |= parse_border(buf,lev+2,off+5+i,&leng,&(stat->top_border));
        i += leng;
        break;
      case 0x12:
        progress(lev+2,off + 4 + i,"Going to read border bottom");
        res |= parse_border(buf,lev+2,off+5+i,&leng,&(stat->bottom_border));
        i += leng;
        break;
      case 0x13:
        progress(lev+2,off + 4 + i,"Going to read border left");
        res |= parse_border(buf,lev+2,off+5+i,&leng,&(stat->left_border));
        i += leng;
        break;
      case 0x14:
        progress(lev+2,off + 4 + i,"Going to read border right");
        res |= parse_border(buf,lev+2,off+5+i,&leng,&(stat->right_border));
        i += leng;
        break;
      case 0x15:
        progress(lev+2,off + 4 + i,"Going to read bullet");
        stat->bullet.on = psiconv_bool_true;
        bul_leng = read_u8(buf,lev+3,off+5+i);
        debug(lev+3,off + 5 + i,"Bullet data length: %02x",bul_leng);
        stat->bullet.font_size = read_point_size(buf,lev+3,off+6+i,&leng);
        debug(lev+3,off+6+i,"Font size: %6.3f",stat->bullet.font_size);
        i += leng;
        stat->bullet.character = read_u8(buf,lev+3,off+6+i);
        debug(lev+3,off+6+i,"Character: %02x",stat->bullet.character);
        temp = read_u8(buf,lev+3,off+7+i);
        if (temp == 0x00)
          stat->bullet.indent = psiconv_bool_false;
        else if (temp == 0x01)
          stat->bullet.indent = psiconv_bool_true;
        else {
          warn(lev+3,off+7+i,
               "Unknown intermediate bullets byte in paragraph layout codes "
               "list");
          debug(lev+3,off+7+i,"Intermediate byte: read %02x",temp);
        }
        stat->bullet.color = read_color_t(buf,lev+3,off+8+i,&leng);
        i += leng;
        res |= parse_font(buf,lev+3,off+8+i,&leng,&(stat->bullet.font),fl);
        i += leng + 3;
        break;
      case 0x16:
        progress(lev+2,off + 5 + i,"Going to read standard tabs");
        stat->tabs.normal.location = read_cm_length(buf,lev+2,off+5 +i,&leng);
        i += leng;
        break;
      case 0x17:
        progress(lev+2,off + 5 + i,"Going to read extra tab");
        tab.location = read_cm_length(buf,lev+3,off+5 +i,&leng);
        temp = read_u8(buf,lev+3,off+i+5 + leng);
        if (temp == 0x01)
          tab.kind = psiconv_tab_left;
        else if (temp == 0x02)
          tab.kind = psiconv_tab_centre;
        else if (temp == 0x03)
          tab.kind = psiconv_tab_right;
        else {
          res = -1;
          warn(lev+3,off+i+5 + leng,"Unknown kind of tab");
          debug(lev+3,off + 5 +i,
                "Tab at %6.3f of unknown type %02x",tab.location,temp);
          i += 1 + leng;
          break;
        }
        debug(lev+3,off + 5 + i,"Tab at %6.3f of type %02x",
              tab.location,tab.kind);
        psiconv_list_add(stat->tabs.extras,&tab);
        i += 1 + leng;
        break;
      default:
        warn(lev+2,off+i,"Unknown code in paragraph layout codes list");
        debug(lev+2,off+ i,"Code: %02x",id);
        res = -1;
        break;
    }
  }

  if (i != nr) {
    warn(lev+2,off+i,
         "Read past end of paragraph layout codes list. I probably lost track"
         "somewhere!");
    debug(lev+2,off+i,"Read %d characters instead of %d",i,nr);
  }

  progress(lev+1,off+nr-3,"End of paragraph layout codes");

  if (len)
    *len = nr + 4;
  return res;
}

int parse_character_layout_codes(const psiconv_buffer buf, int lev,
                                 psiconv_u32 off, int *len, 
                                 psiconv_character_layout_status stat, 
                                 psiconv_font_list fl)
{
  int res=0;
  int nr,i,leng;
  psiconv_u8 id;
  psiconv_u32 temp;

  progress(lev+1,off,"Going to read character layout codes");
  nr = read_u32(buf,lev+2,off);
  debug(lev+2,off,"Length: %08x",nr);
  for (i = 0; i < nr; i++) {
    id = read_u8(buf,lev+2,off + 4 + i);
    switch(id) {
      case 0x19:
        progress(lev+2,off+4+i,"Going to read color");
        stat->color = read_color_t(buf,lev+2,off + 5 + i, &leng);
        i += leng;
        break;
      case 0x1a:
        progress(lev+2,off+4+i,"Going to read background color (?)");
        stat->back_color = read_color_t(buf,lev+2,off + 5 + i, &leng);
        i += leng;
        break;
      case 0x1c:
        progress(lev+2,off+4+i,"Going to read font size");
        stat->font_size = read_point_size(buf,lev+2,off+5+i,&leng);
        i += leng;
        debug(lev+2,off+5+i,"Size: %6.3f",stat->font_size);
        break;
      case 0x1d:
        progress(lev+2,off+4+i,"Going to read italic");
        temp = read_u8(buf,lev+2,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->italic = psiconv_to_bool(temp);
        else {
          warn(lev+2,off+5+i,
               "Unknown italic argument in character layout codes list");
          res = -1;
        }
        debug(lev+2,off+5+i,"Italic: %02x",temp);
        i ++;
        break;
      case 0x1e:
        progress(lev+2,off+4+i,"Going to read bold");
        temp = read_u8(buf,lev+2,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->bold = psiconv_to_bool(temp);
        else {
          warn(lev+2,off+5+i,
               "Unknown bold argument in character layout codes list");
          res = -1;
        }
        debug(lev+2,off+5+i,"Bold: %02x",temp);
        i ++;
        break;
      case 0x1f:
        progress(lev+2,off+4+i,"Going to read super_sub");
        temp = read_u8(buf,lev+2,off + 5 + i);
        if (temp == 0x00)
          stat->super_sub = psiconv_normalscript;
        else if (temp == 0x01)
          stat->super_sub = psiconv_superscript;
        else if (temp == 0x02)
          stat->super_sub = psiconv_subscript;
        else {
          warn(lev+2,off+5+i,
               "Unknown super_sub argument in character layout codes list");
          res = -1;
        }
        debug(lev+2,off+5+i,"Super_sub: %02x",temp);
        i ++;
        break;
      case 0x20:
        progress(lev+2,off+4+i,"Going to read underline");
        temp = read_u8(buf,lev+2,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->underline = psiconv_to_bool(temp);
        else {
          warn(lev+2,off+5+i,
               "Unknown underline argument in character layout codes list");
          res = -1;
        }
        debug(lev+2,off+5+i,"Underline: %02x",temp);
        i ++;
        break;
      case 0x21:
        progress(lev+2,off+4+i,"Going to read crossed_out");
        temp = read_u8(buf,lev+2,off + 5 + i);
        if (temp == 0x00 || temp == 0x01)
          stat->crossed_out = psiconv_to_bool(temp);
        else {
          warn(lev+2,off+5+i,
               "Unknown crossed_out argument in character layout codes list");
          res = -1;
        }
        debug(lev+2,off+5+i,"Crossed_out: %02x",temp);
        i ++;
        break;
      case 0x22:
        progress(lev+2,off+4+i,"Going to read font");
        res |= parse_font(buf,lev+2,off + 5 + i, &leng, &stat->font,fl);
        debug(lev+2,off+5+i,"Font: %02x",stat->font);
        i += leng;
        break;
      default:
        warn(lev+2,off+i,"Unknown code in character layout codes list");
        debug(lev+2,off+i,"Code: %02x",id);
        res = -1;
        break;
    }
  }

  if (i != nr) {
    warn(lev+2,off+i,
         "Read past end of character layout codes list. I probably lost track"
         "somewhere!");
    debug(lev+2,off+i,"Read %d characters instead of %d",i,nr);
  }

  if (len)
    *len = nr + 4;

  progress(lev+1,off+nr-3,"End of character layout codes");

  return res;
}

int parse_styles_section(const psiconv_buffer buf, int lev, psiconv_u32 off, 
                         int *length,psiconv_styles_section *ss, 
                         psiconv_font_list fl)
{
  int res = 0;
  int len=0;
  int leng,i,nr;
  psiconv_style s,this_style;
  psiconv_u32 temp;
  
  progress(lev+1,off+len,"Going to read the styles section");
  (*ss) = malloc(sizeof(**ss));

  progress(lev+2,off+len,"Going to read style normal");
  s = malloc(sizeof(*s));
  s->name = NULL;
  s->unknown = 0x00000000;
  s->paragraph = psiconv_basic_paragraph_layout_status();
  progress(lev+3,off+len,"Going to read the paragraph codes");
  res |= parse_paragraph_layout_codes(buf,lev+3,off+len,&leng, s->paragraph,fl);
  len += leng;
  progress(lev+3,off+len,"Going to read the character codes");
  s->character = psiconv_basic_character_layout_status();
  res |= parse_character_layout_codes(buf,lev+3,off+len,&leng, s->character,fl);
  len += leng;
  progress(lev+3,off+len,"Going to read the hotkey");
  temp = read_u32(buf,lev+3,off+len);
  if (temp & 0xffffff00) {
    warn(lev+3,off+len,"Normal style hotkey has unknown value");
    debug(lev+3,off+len,"Hotkey value %08x",temp);
    temp = 0;
    res = -1;
  } 
  s->hotkey = temp;
  if ((s->hotkey >= 32) && (s->hotkey < 127))
    debug(lev+3,off+len,"Hotkey: %c",s->hotkey);
  else if (s->hotkey == 0x00)
    debug(lev+3,off+len,"Hotkey: <none>");
  else
    debug(lev+3,off+len,"Hotkey: %02x",s->hotkey);
  (*ss)->normal = s;
  len += 0x04;

  progress(lev+2,off+len,"Going to read hotkeys list");
  (*ss)->styles = psiconv_list_new(sizeof(*s));
  s = malloc(sizeof(*s));
  progress(lev+3,off+len,"Going to read the number of entries");
  nr = read_u8(buf,lev+3,off+len);
  len ++;
  debug(lev+3,off+len,"Nummer of hotkeys: %02x",nr);
  for (i = 0; i < nr; i ++) {
    temp = read_u32(buf,lev+2,off+len);
    if (temp & 0xffffff00) {
      warn(lev+3,off+len,"Style hotkey has unknown value");
      debug(lev+3,off+len,"Hotkey %d value %08x",i,temp);
      temp = 0;
      res = -1;
    } 
    s->hotkey = temp;
    if ((s->hotkey >= 32) && (s->hotkey < 127))
      debug(lev+3,off+len,"Hotkey %d: %c",i,s->hotkey);
    else if (s->hotkey == 0x00)
      debug(lev+3,off+len,"Hotkey %d: <none>",i);
    else
      debug(lev+3,off+len,"Hotkey %d: %02x",i,s->hotkey);
    psiconv_list_add((*ss)->styles,s);
    len += 0x04;
  }
  free(s);

  progress(lev+2,off+len,"Going to read other styles");
  progress(lev+3,off+len,"Going to read the number of styles");
  nr = read_u8(buf,lev+3,off+len);
  if (nr != psiconv_list_length((*ss)->styles)) {
    warn(lev+3,off+len,"Number of styles and hotkeys do not match");
    debug(lev+3,off+len,"%d hotkeys, %d styles",
          psiconv_list_length((*ss)->styles), nr);
  }
  len ++;

  for (i = 0; i < nr; i++) {
    if (i >= psiconv_list_length((*ss)->styles)) {
      s->hotkey = 0;
      psiconv_list_add((*ss)->styles,s);
      debug(lev+3,off+len,"Style %d: new entry added in list",i);
    }
    this_style = psiconv_list_get((*ss)->styles,i);
    progress(lev+3,off+len,"Style %d: Going to read the style name",i);
    this_style->name = read_string(buf,lev+2,off+len,&leng);
    len += leng;
    progress(lev+3,off+len,"Style %d: Going to read an unknown id",i);
    this_style->unknown = read_u32(buf,lev+2,off+len);
    debug(lev+2,off+len,"Style %d unknown id: %08x",i,this_style->unknown);
    len += 4;
    progress(lev+3,off+len,"Style %d: Going to read an unknown offset (?) "
                           "(0x%08x expected)",i,0x00000000);
    temp = read_u32(buf,lev+3,off+len);
    if (temp != 0x00000000) {
      warn(lev+3,off+len,"Unexpected value in style description");
      debug(lev+3,off+len,"Style %d long: read %08x, expected %08x",i,temp,
            0x00000000);
    }
    len += 4;
    progress(lev+3,off+len,"Style %d: Going to read the character codes",i);
    this_style->character = psiconv_basic_character_layout_status();
    res |= parse_character_layout_codes(buf,lev+2,off+len,&leng,
                                        this_style->character,fl);
    len += leng;
    progress(lev+3,off+len,"Style %d: Going to read the paragraph codes",i);
    this_style->paragraph = psiconv_basic_paragraph_layout_status();
    res |= parse_paragraph_layout_codes(buf,lev+2,off+len,&leng,
                                        this_style->paragraph,fl);
    len += leng;
  }

  progress(lev+1,off+len-1,"End of section table");

  if (length)
    *length = len;

  return res;
}


void no_layout_section(psiconv_text_and_layout t,psiconv_styles_section ss)
{
  psiconv_paragraph para;
  int i;

  for (i = 0; i < psiconv_list_length(t); i ++) {
    para = psiconv_list_get(t,i);
    para->in_lines = psiconv_list_new(sizeof(struct psiconv_in_line_layout));
    para->base_style = 0;
    para->base_character = 
                 psiconv_clone_character_layout_status(ss->normal->character);
    para->base_paragraph = 
                 psiconv_clone_paragraph_layout_status(ss->normal->paragraph);
  }
}

/* First do a parse_text_section, or you will get into trouble here */
int parse_layout_section(const psiconv_buffer buf,int lev,psiconv_u32 off,
                         int *length,psiconv_text_and_layout t,
                         psiconv_font_list fl,psiconv_styles_section ss)
{
  int res = 0;
  int len = 0;
  int leng,nr,i,j,k,l;
  psiconv_u32 temp;

  typedef struct anon_style
  {
    int nr;
    psiconv_s16 base_style;
    psiconv_character_layout_status character;
    psiconv_paragraph_layout_status paragraph;
  } *anon_style;

  typedef psiconv_list anon_style_list; /* of anon_style */

  anon_style_list anon_styles;
  struct anon_style this_anon;
  psiconv_in_line_layout this_in_line;
  anon_style anon_ptr = NULL;
  psiconv_u32 *nr_els;
  psiconv_paragraph par;
  int sectype; /* 0 for style-less, 1 for with styles, -1 for ignore styles */

  progress(lev+1,off,"Going to read the layout section");
  progress(lev+2,off,"Going to read the section type");
  temp = read_u16(buf,lev+2,off+len);
  debug(lev+2,off+len,"Type: %02x",temp);
  if (temp == 0x0001) {
    if (!ss) {
      debug(lev+2,off+len,"Type: read %08x, expected %08x",temp,0x0001);
      warn(lev+2,off+len,"Type of the layout section mismatch. "
                     "I expected a style-less layout section, but "
                     "this is not it! Ignoring style markers...");
      sectype = -1;
    } else
      sectype = 1;
  } else {
    sectype = 0;
    if (ss) {
      debug(lev+2,off+len,"Type: read %08x, expected %08x",temp,0x0000);
      warn(lev+2,off+len,"Type of the layout section mismatch. "
                     "I expected a styled layout section, but "
                     "this is not it!  style markers...");
      warn(lev+2,off+len,"First word read of the preamble of the layout section "
                     "failed. I expected a styled layout section, but "
                     "this is not it! Ignoring all styles...");
    }
  }
  len = 0x02;

  progress(lev+2,off+len,"Going to read paragraph type list");
  anon_styles = psiconv_list_new(sizeof(this_anon));
  progress(lev+2,off+len,"Going to read paragraph type list length");
  nr = read_u8(buf,lev+2,off + len);
  debug(lev+2,off+len,"Length: %02x",nr);
  len ++;

  progress(lev+3,off+len,"Going to read paragraph type list elements");
  for (i = 0; i < nr; i ++) {
    this_anon.nr = read_u32(buf,lev+3,off + len);
    debug(lev+3,off+len,"Element %d Type: %08x",i,this_anon.nr);
    len += 0x04;
    if (sectype <= 0) {
      this_anon.base_style = -1;
      this_anon.paragraph = psiconv_basic_paragraph_layout_status();
      this_anon.character = psiconv_basic_character_layout_status();
    } else if (sectype == 1) { 
      this_anon.base_style = read_u8(buf, lev+3,
                                  off + len +0x04 +read_u32(buf,lev+3,off+len));
      this_anon.paragraph = psiconv_clone_paragraph_layout_status(
                        psiconv_get_style(ss,this_anon.base_style)->paragraph);
      this_anon.character = psiconv_clone_character_layout_status(
                        psiconv_get_style(ss,this_anon.base_style)->character);
    }
    debug(lev+3,off+len,"Element %d Base style: %02x",i,this_anon.base_style);

    res |= parse_paragraph_layout_codes(buf,lev+3,off+len,&leng,
                                        this_anon.paragraph, fl);
    debug(lev+3,off+len,"Element %d Base style: %02x",i,this_anon.base_style);
    len += leng;
    if (sectype)
      len ++;
    res |= parse_character_layout_codes(buf,lev+3,off+len,&leng,
                                        this_anon.character, fl);
    psiconv_list_add(anon_styles,&this_anon);
    len += leng;
  }

  progress(lev+2,off+len,"Going to parse the paragraph element sequence");
  progress(lev+3,off+len,"Going to read the number of paragraphs");
  nr = read_u32(buf,lev+3,off+len);
  if (nr != psiconv_list_length(t)) {
    warn(lev+3,off+len,
         "Number of text paragraphs and paragraph elements does not match");
    debug(lev+3,off+len,
          "%d text paragraphs, %d paragraph elements",
          psiconv_list_length(t),nr);
  } 

  len += 4;

  nr_els = malloc(nr * sizeof(*nr_els));
  progress(lev+3,off+len,"Going to read the paragraph elements");
  for (i = 0; i < nr; i ++) {
    if (i >= psiconv_list_length(t)) {
      par = malloc(sizeof(*par));
      par->text = strdup("");
      psiconv_list_add(t,par);
      free(par);
      res = -1;
      debug(lev+3,off+len,"Paragraph %d: new entry added in list",i);
    }
    par = psiconv_list_get(t,i);
    
    temp = read_u32(buf,lev+3,off+len);
    if (temp != strlen(par->text)+1) {
      res = -1;
      warn(lev+3,off+len,
           "Disagreement of the length of paragraph %d in layout section",i);
      debug(lev+3,off+len,
            "Paragraph %d length: layout section says %d, counted %d",i,
            temp,strlen(par->text)+1);
    } else
      debug(lev+3,off+len,"Paragraph %d length: %d",i,temp);
    
    temp = read_u8(buf,lev+3,off + len + 0x04);
    if (temp != 0x00) {
      debug(lev+3,off+len+0x04,"Paragraph %d type: %02x",i,temp);
      for (j = 0; j < psiconv_list_length(anon_styles); j ++) {
        anon_ptr = psiconv_list_get(anon_styles,j);
        if (temp == anon_ptr->nr)
          break;
      }
      if (j == psiconv_list_length(anon_styles)) {
        warn(lev+3,off+len+0x04,"Unknown paragraph type for paragraph %d",i);
        debug(lev+3,off+len+0x04,
              "Paragraph %d unknown type %02x - falling back on base style",
              i,temp);
        par->base_style = 0;
        par->base_character = psiconv_basic_character_layout_status();
        par->base_paragraph = psiconv_basic_paragraph_layout_status();
      } else {
        par->base_style = anon_ptr->base_style;
        par->base_character = psiconv_clone_character_layout_status
                                                       (anon_ptr->character);
        par->base_paragraph = psiconv_clone_paragraph_layout_status
                                                       (anon_ptr->paragraph);
      }
      nr_els[i] = 0;
      len += 5;
    } else {
      debug(lev+3,off+len,"Paragraph %d: not based on a paragraph type",i);

    if (sectype <= 0) {
      par->base_style = -1;
      par->base_paragraph = psiconv_basic_paragraph_layout_status();
      par->base_character = psiconv_basic_character_layout_status();
    } else if (sectype == 1) {
      par->base_style = read_u8(buf, lev+3,
                             off + len +0x09 +read_u32(buf,lev+3,off+len+0x05));
      par->base_paragraph = psiconv_clone_paragraph_layout_status(
                        psiconv_get_style(ss,par->base_style)->paragraph);
      par->base_character = psiconv_clone_character_layout_status(
                        psiconv_get_style(ss,par->base_style)->character);
    }

    debug(lev+3,off+len+0x05,
            "Paragraph %d Base style: %02x",i,par->base_style);
    res |= parse_paragraph_layout_codes(buf,lev+3,off + len + 0x05,&leng,
                                        par->base_paragraph,fl);
    if (sectype)
      len ++;

    nr_els[i] = read_u32(buf,lev+3,off + len + leng + 0x05);
    debug(lev+3,off + len + leng + 0x05,
            "Paragraph %d: %d layout elements",i,nr_els[i]);
      len += leng + 0x09;
    }
  }

  progress(lev+2,off,"Going to read the layout element sequence");
  progress(lev+3,off,"Going to read the number of elements");
  nr = read_u32(buf,lev+3,off+len);
  debug(lev+3,off,"Elements: %08x",nr);
  len += 0x04;

  k = 0;
  this_in_line = malloc(sizeof(*this_in_line));
  for (i = 0; i < psiconv_list_length(t); i++) {
    l = 0;
    par = psiconv_list_get(t,i);
    par->in_lines = psiconv_list_new(sizeof(*this_in_line));
    progress(lev+3,off+len,
             "Paragraph %d: %d elements (current: %d, left: %d)",i,nr_els[i],
             k, nr - k);
    this_in_line->layout = par->base_character;
    if (k + nr_els[i] > nr) {
      res = -1;
      warn(lev+3,off+len,
           "Paragraph %d: layout elements list exhausted, can't register "
           "internal layout");
    }
    for (j = k; (j < nr) && (j < k + nr_els[i]); j++) {
      this_in_line->layout = psiconv_clone_character_layout_status
                                                       (this_in_line->layout);
      temp = read_u8(buf,lev+3,off+len);
      debug(lev+3,off+len,"Paragraph %d entry %d type: %02x",i,j,temp);
      this_in_line->length = read_u32(buf,lev+3,off+len+1);
      debug(lev+3,off+len+1,
            "Paragraph %d entry %d length: %08x",i,j,this_in_line->length);
      l += this_in_line->length;
      this_in_line->layout->italic = par->base_character->italic;
      this_in_line->layout->bold = par->base_character->bold;
      this_in_line->layout->super_sub = par->base_character->super_sub;
      this_in_line->layout->underline = par->base_character->underline;
      this_in_line->layout->crossed_out = par->base_character->crossed_out;
      res |= parse_character_layout_codes(buf,lev+3,off + len + 5, &leng,
                                          this_in_line->layout,fl);
      if (temp == 0x01) {
        debug(lev+3,off+len+5+leng,
              "Paragraph %d entry %d: skipped object data",i,j);
        len += 0x10;
      } else if (temp != 0x00) {
        debug(lev+3,off+len,
              "Paragraph %d entry %d: unknown type. I may get confused now.");
        nr = -1;
      }
      psiconv_list_add(par->in_lines,this_in_line);
      len += 5 + leng;
    }
    k = j;
  }
  if (k != nr) {
    res = -1;
    warn(lev+3,off+len,"Some unused layout elements are left");
    debug(lev+3,off+len,"Elements: Paragraphs wanted %d, we found %d",nr,k);
  }

  progress(lev,off+len-1,"End of layout section");

  if (length)
    *length = len;

  free(this_in_line);

  return res;
}

int parse_header_footer(const psiconv_buffer buf,int lev,psiconv_u32 off,
                        int *length,psiconv_header *hd,psiconv_font_list fl)
{
  int res = 0;
  int len = 0;
  int i,leng;
  psiconv_u32 subsecoff,layoutsecoff,temp;
  psiconv_paragraph para;

  
  progress(lev+1,off,"Going to read a header or footer");
  (*hd) = malloc(sizeof(**hd));

  progress(lev+2,off+len,"Going to read the first byte");
  temp = read_u8(buf,lev+2,off+len);
  if (temp != 0x01) {
    warn(lev+2,off+len,"Header/footer first byte mismatch");
    debug(lev+2,off+len,"First byte: read %02x, expected %02x",temp,0x01);
    res = -1;
  }
  len += 1;

  progress(lev+2,off+len,"Going to read displayed-on-first-page");
  temp = read_u8(buf,lev+2,off+len);
  if (temp == 0x00 || temp == 0x01) 
    (*hd)->on_first_page = psiconv_to_bool(temp);
  else {
    warn(lev+2,off+len,"Header/footer unknown value for "
                       "displayed-on-first-page byte");
    (*hd)->on_first_page = psiconv_bool_true;
    res = -1;
  }
  debug(lev+2,off+len,"Displayed-on-first-page: %02x",temp);
  len += 1;

  progress(lev+2,off+len,"Going to read intermediate section of zero bytes");
  for (i = 0; i < 0x03; i++) {
    temp = read_u8(buf,lev+2,off+len+i);
    if (temp != 0x00) {
      warn(lev+2,off+len+i,
           "Header/footer unknown value in intermediate section");
      debug(lev+2,off+len+i,"Byte %d: read %02x, expected %02x",i,temp,0x00);
      res = -1;
    }
  }
  len += 0x03;
  
  progress(lev+2,off+len,"Going to discard paragraph layout section");
  temp = read_u32(buf,lev+2,off+len);
  len += 4;
  for (i=0; i<temp; i++) read_u8(buf,lev+2,off+len+i);
  len += temp;

  progress(lev+2,off+len,"Going to discard character layout section");
  temp = read_u32(buf,lev+2,off+len);
  len += 4;
  for (i=0; i<temp; i++) read_u8(buf,lev+2,off+len+i);
  len += temp;

  progress(lev+2,off+len,"Going to read first identifier 0x1000005c");
  temp = read_u32(buf,lev+2,off+len);
  if (temp != 0x1000005C) {
    warn(lev+2,off+len,"Header/footer identifier mismatch");
    debug(lev+2,off+len,"Identifier: read %08x, expected %08x",temp,0x1000005c);
    res = -1;
  }
  len += 4;

  progress(lev+2,off+len,"Going to read identifier 0x10000063");
  temp = read_u32(buf,lev+2,off+len);
  if (temp != 0x10000063) {
    warn(lev+2,off+len,"Header/footer second identifier mismatch");
    debug(lev+2,off+len,"Identifier: read %08x, expected %08x",temp,0x10000063);
    res = -1;
  }
  len += 4;

  progress(lev+2,off+len,"Going to read the subsection offset");
  subsecoff = read_u32(buf,lev+2,off+len);
  debug(lev+2,off+len,"Subsection offset: %08x",subsecoff);
  len += 4;
  
  progress(lev+2,off+len,"Going to read identifier 0x10000065");
  temp = read_u32(buf,lev+2,off+len);
  if (temp != 0x10000065) {
    warn(lev+2,off+len,"Header/footer third identifier mismatch");
    debug(lev+2,off+len,"Identifier: read %08x, expected %08x",temp,0x10000065);
    res = -1;
  }
  len += 4;

  progress(lev+2,off+len,"Going to read the (unknown) offset "
                         "(0x00000000 expected)");
  temp = read_u32(buf,lev+2,off+len);
  if (temp != 0x00000000) {
    warn(lev+2,off+len,"Header/footer unknown section offset found");
    debug(lev+2,off+len,"Offset: read %08x, expected %08x",temp,0x00000000);
    res = -1;
  }
  len += 0x04;
  
  progress(lev+2,off+len,"Going to read identifier 0x10000066");
  temp = read_u32(buf,lev+2,off+len);
  if (temp != 0x10000066) {
    warn(lev+2,off+len,"Header/footer fourth identifier mismatch");
    debug(lev+2,off+len,"Identifier: read %08x, expected %08x",temp,0x10000066);
    res = -1;
  }
  len += 4;

  progress(lev+2,off+len,"Going to read the layout section offset");
  layoutsecoff = read_u32(buf,lev+2,off+len);
  debug(lev+2,off+len,"Layout section offset: %08x",layoutsecoff);
  len += 0x04;
  
  progress(lev+2,off+len,"Going to read identifier 0x10000064");
  temp = read_u32(buf,lev+2,off+len);
  if (temp != 0x10000064) {
    warn(lev+2,off+len,"Header/footer fifth identifier mismatch");
    debug(lev+2,off+len,"Identifier: read %08x, expected %08x",temp,0x10000064);
    res = -1;
  }
  len += 0x04;

  progress(lev+2,off+len,"Going to read the text section");
  res |= parse_text_section(buf,lev+2,off+len,&leng,&(*hd)->paragraphs);
  len += leng;
 
  progress(lev+2,off+len,"Going to read the style-less layout section");
  if (layoutsecoff) 
    res |= parse_layout_section(buf,lev+2,layoutsecoff,NULL,(*hd)->paragraphs,
                                fl,NULL);
  else 
    for (i = 0;i < psiconv_list_length((*hd)->paragraphs); i ++) {
      para = psiconv_list_get((*hd)->paragraphs,i);
      para->in_lines = psiconv_list_new(sizeof(struct psiconv_in_line_layout));
      para->base_style = -1;
      para->base_character = 
                 psiconv_basic_character_layout_status();
      para->base_paragraph = 
                 psiconv_basic_paragraph_layout_status();
    }

  progress(lev+1,off+len-1,"End of header or footer");

  if (length)
    *length = len;
  return res;
}

int parse_page_section(const psiconv_buffer buf,int lev,psiconv_u32 off,
                       int *length, psiconv_page_section *ps, 
                       psiconv_font_list fl)
{
  int res = 0;
  int len = 0;
  int leng;
  psiconv_u32 temp;

  progress(lev+1,off,"Going to read the page section");
  (*ps) = malloc(sizeof(**ps));

  progress(lev+2,off+len,"Going to read first page number");
  (*ps)->first_page_nr = read_u32(buf,lev+2,off+len);
  debug(lev+2,off+len,"First page: %d",(*ps)->first_page_nr);
  len += 4;
  
  progress(lev+2,off+len,"Going to read header distance");
  (*ps)->header_dist = read_cm_length(buf,lev+2,off+len,&leng);
  debug(lev+2,off+len,"Header distance: %6.3f",(*ps)->header_dist);
  len += leng;
  
  progress(lev+2,off+len,"Going to read footer distance");
  (*ps)->footer_dist = read_cm_length(buf,lev+2,off+len,&leng);
  debug(lev+2,off+len,"Footer distance: %6.3f",(*ps)->footer_dist);
  len += leng;
  
  progress(lev+2,off+len,"Going to read the left margin");
  (*ps)->left_margin = read_cm_length(buf,lev+2,off+len,&leng);
  debug(lev+2,off+len,"Left margin: %6.3f",(*ps)->left_margin);
  len += leng;
  
  progress(lev+2,off+len,"Going read the to right margin");
  (*ps)->right_margin = read_cm_length(buf,lev+2,off+len,&leng);
  debug(lev+2,off+len,"Right margin: %6.3f",(*ps)->right_margin);
  len += leng;
  
  progress(lev+2,off+len,"Going to read the top margin");
  (*ps)->top_margin = read_cm_length(buf,lev+2,off+len,&leng);
  debug(lev+2,off+len,"Top margin: %6.3f",(*ps)->top_margin);
  len += leng;
  
  progress(lev+2,off+len,"Going to read the bottom margin");
  (*ps)->bottom_margin = read_cm_length(buf,lev+2,off+len,&leng);
  debug(lev+2,off+len,"Bottom margin: %6.3f",(*ps)->bottom_margin);
  len += leng;

  progress(lev+2,off+len,"Going to read the header");
  res |= parse_header_footer(buf,lev+2,off+len,&leng,&(*ps)->header,fl);
  len += leng;
  
  progress(lev+2,off+len,"Going to read the footer");
  res |= parse_header_footer(buf,lev+2,off+len,&leng,&(*ps)->footer,fl);
  len += leng;

  progress(lev+2,off+len,"Going to read the page size marker");
  temp = read_u32(buf,lev+2,off+len);
  if (temp != 0x100000FD) {
    warn(lev+2,off+len,"Page section page size marker not found");
    debug(lev+2,off+len,"Page size marker: read %08x, expected %08x",temp,
          0x100000FD);
    res = -1;
  }
  
  progress(lev+2,off+len,"Going to read the page width");
  (*ps)->page_width = read_cm_length(buf,lev+2,off+len,&leng);
  debug(lev+2,off+len,"Page width: %6.3f",(*ps)->page_width);
  len += leng;

  progress(lev+2,off+len,"Going to read the page height");
  (*ps)->page_height = read_cm_length(buf,lev+2,off+len,&leng);
  debug(lev+2,off+len,"Page width: %6.3f",(*ps)->page_height);
  len += leng;

  progress(lev,off+len-1,"End of page section");

  if (length)
    *length = len;

  return res;
}


psiconv_s64 offset_in_sectiontable(psiconv_u32 id,
                                   const psiconv_sectiontable_section s)
{
  int i;
  psiconv_sectiontable_entry ep;
  for (i = 0; i < psiconv_list_length(s); i++) {
    ep = psiconv_list_get(s,i),id;
    if (id == ep->id)
      return ep->offset;
  }
  return -1;
}


int psiconv_parse(const psiconv_buffer buf,psiconv_word_file *wf)
{

  psiconv_header_section h;
  psiconv_sectiontable_section s;
  psiconv_appl_id_section a;
  int off;

  int res = 0;
  int lev = 0;
  *wf = malloc(sizeof(**wf));

  res = parse_header_section(buf,lev+1,0,NULL,&h);
  res |= parse_sectiontable(buf,lev+1,h->sectiontable_offset,NULL,&s);

  progress(lev,h->sectiontable_offset,"Looking for the Status section");
  off = offset_in_sectiontable(PSICONV_ID_STATUS_SECTION,s);
  if (off < 0) {
   warn(lev,h->sectiontable_offset,
        "Status section not found in the section table");
    res = -1;
  } else {
    debug(lev,h->sectiontable_offset,"Status section at offset %08x",off);
    res |= parse_status_section(buf,lev+1,off,NULL,&((*wf)->status_sec));
  }

  progress(lev,h->sectiontable_offset,"Looking for the Application ID section");
  off = offset_in_sectiontable(PSICONV_ID_APPL_ID_SECTION,s);
  if (off < 0) {
    warn(lev,h->sectiontable_offset,
         "Application ID section not found in the section table");
    res = -1;
  } else {
    debug(lev,h->sectiontable_offset,
          "Application ID section at offset %08x",off);
    res |= parse_appl_id_section(buf,lev+1,off,NULL,&a);
  }

  (*wf)->fonts = psiconv_list_new(sizeof(struct psiconv_font_desc));

  progress(lev,h->sectiontable_offset,"Looking for the Page section");
  off = offset_in_sectiontable(PSICONV_ID_PAGE_SECTION,s);
  if (off < 0) {
    warn(lev,h->sectiontable_offset,
         "Page section not found in the section table");
    res = -1;
  } else {
    debug(lev,h->sectiontable_offset,"Page section at offset %08x",off);
    res |= parse_page_section(buf,lev+1,off,
                              NULL,  &((*wf)->page_sec),(*wf)->fonts);
  }

  progress(lev,h->sectiontable_offset,"Looking for the Styles section");
  off = offset_in_sectiontable(PSICONV_ID_STYLES_SECTION,s);
  if (off < 0) {
    warn(lev,h->sectiontable_offset,
         "Styles section not found in the section table");
    res = -1;
  } else {
    debug(lev,h->sectiontable_offset,"Styles section at offset %08x",off);
    res |= parse_styles_section(buf,lev+1,off,
                                NULL,&((*wf)->styles_sec),(*wf)->fonts);
  }

  progress(lev,h->sectiontable_offset,"Looking for the Text section");
  off = offset_in_sectiontable(PSICONV_ID_TEXT_SECTION,s);
  if (off < 0) {
    warn(lev,h->sectiontable_offset,
         "Text section not found in the section table");
    res = -1;
  } else {
    debug(lev,h->sectiontable_offset,"Text section at offset %08x",off);
    res |= parse_text_section(buf,lev+1,off,NULL,&((*wf)->paragraphs));
  }

  progress(lev,h->sectiontable_offset,
           "Looking for the (optional) Layout section");
  off = offset_in_sectiontable(PSICONV_ID_LAYOUT_SECTION,s);
  if (off < 0) {
    debug(lev,h->sectiontable_offset,
          "Layout section not found in the section table");
    no_layout_section((*wf)->paragraphs,(*wf)->styles_sec);
  } else {
    debug(lev,h->sectiontable_offset,"Layout section at offset %08x",off);
    res |= parse_layout_section(buf,lev,off,NULL,(*wf)->paragraphs,(*wf)->fonts,
                                (*wf)->styles_sec);
  }
  return res;
}
